"""
Main module for reading maml files.
"""
import warnings
import os

import yaml
from yaml.parser import ParserError

warnings.formatwarning = lambda msg, *args, **kwargs: f"{msg}\n"

def read_maml(file_name: str) -> dict:
    """
    Reads in a maml file and warns for strange extensions.
    """
    _, extension = os.path.splitext(file_name)
    match extension:
        case '.yml' | '.yaml':
            warnings.warn("WARNING: File was read in but this is a .yml not a .maml\n")
        case '.maml':
            pass
        case _:
            warnings.warn("WARNING: Attempting to read in file but extension is not valid.")
    with open(file_name, encoding='utf8') as file:
        try:
            maml_dict = yaml.safe_load(file)
        except ParserError as exc:
            raise ValueError("File is not even valid YAML. Failed to read.") from exc
    if not maml_dict:
        warnings.warn("FILE IS EMPTY!")
        maml_dict = {}
    return maml_dict



# src/pymaml/__init__.py

from .maml import MAML
from .parse import is_iso8601, valid_for
from .read import read_maml
from .model_v1p0 import V1P0
from .model_v1p1 import V1P1

__all__ = [
    "MAML", 
    "read_maml",
    "is_iso8601",
    "V1P0",
    "V1P1",
    "valid_for",
]



"""
Pydantic base model for MAML version 1.1
"""

# generated by datamodel-codegen:
# filename:  v1p1.json
# timestamp: 2025-09-16T01:55:16+00:00
# datamodel-codegen --input schemas/v1p1.json --input-file-type jsonschema --output model_v1p1.py

from __future__ import annotations

from datetime import date as date_aliased
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field
from astropy.io.votable.ucd import check_ucd


class DOIEntry(BaseModel):
    """
    DOI field that holds the DOI and the type of DOI.
    """

    DOI: str = Field(..., description="Valid DOI")
    type: str = Field(..., description="Type of DOI")


class DependEntry(BaseModel):
    """
    Depends field that tells you what tables the meta data depends on.
    """

    survey: Optional[str] = Field(None, description="The name of the dependent survey.")
    dataset: Optional[str] = Field(
        None, description="The name of the dependent dataset."
    )
    table: str = Field(..., description="The name of the dependent table.")
    version: Optional[Union[str, float]] = Field(
        None, description="The version of the dependent table."
    )


class KeyValueEntry(BaseModel):
    """
    Allowing Key-value entries into the meta data. Main change from v1.0
    """

    key: str = Field(..., description="Key name")
    value: Union[Union[float, str], List[Union[float, str]]] = Field(
        ..., description="Value/s of key"
    )
    comment: str = Field(..., description="Description of key")


class QCEntry(BaseModel):
    """
    Quality Control field that allows for min/max checks during data validation.
    """

    min: Optional[Union[float, str]] = None
    max: Optional[Union[float, str]] = None
    miss: Optional[Union[float, str]] = None


class FieldEntry(BaseModel):
    """
    Fields field describing the columns of the table.
    """

    name: str = Field(..., description="Required field name")
    unit: Optional[str] = Field(None, description="Recommended unit of measurement")
    info: Optional[str] = Field(None, description="Recommended short description")
    ucd: Optional[Union[str, List[str]]] = Field(
        None, description="Recommended UCD string"
    )
    data_type: str = Field(
        ..., description="Required data type (e.g., int32, string, bool, double)"
    )
    array_size: Optional[Union[int, str]] = Field(
        None, description="Optional max length of string"
    )
    qc: Optional[QCEntry] = Field(
        None, description="Optional quality control parameters"
    )

    def __post_init__(self):
        """
        Checking that the ucd, if it exists, is valid.
        """
        if self.ucd is not None:
            check_ucd(self.ucd)


class V1P1(BaseModel):
    """Pydantic Base model for version 1.0"""

    class Config:
        """Setting configuration. Not allowing extra fields"""

        extra = "forbid"

    survey: Optional[str] = Field(None, description="Optional survey name")
    dataset: Optional[str] = Field(None, description="Recommended dataset name")
    table: str = Field(..., description="Required table name")
    version: Union[str, float] = Field(
        ..., description="Required version (string, integer, or float)"
    )
    date: date_aliased = Field(
        ..., description="Required date in YYYY-MM-DD format (ISO-8601)"
    )
    author: str = Field(..., description="Required lead author name and <email>")
    coauthors: Optional[Union[str, List[str]]] = Field(
        None, description="Optional coauthor name and optionally <email>"
    )
    DOIs: Optional[List[DOIEntry]] = None
    depends: Optional[List[DependEntry]] = None
    description: Optional[str] = Field(
        None, description="Recommended short description of the table"
    )
    comments: Optional[Union[str, List[str]]] = Field(
        None, description="Optional comment or interesting fact"
    )
    license: Optional[str] = Field(None, description="Recommended license for the data")
    keywords: Optional[Union[str, List[str]]] = Field(
        None, description="Optional keyword tag"
    )
    keyarray: Optional[List[KeyValueEntry]] = None
    extra: Optional[Dict[str, Any]] = None
    MAML_version: float = Field(
        1.1, literal=True, description="Optional version of the MAML schema"
    )
    fields: List[FieldEntry]



"""
Main maml object.
"""

import os

import yaml
from yaml import SafeDumper


from .read import read_maml
from .model_v1p0 import V1P0
from .model_v1p1 import V1P1
from .parse import MODELS, _assert_version



def _remove_nones(obj):
    if isinstance(obj, dict):
        return {
            k: _remove_nones(v)
            for k, v in obj.items()
            if v is not None
        }
    if isinstance(obj, list):
        return [_remove_nones(v) for v in obj if v is not None]
    return obj

class MAML:
    """
    Main MAML class
    """

    def __init__(self, data: dict, version):
        """
        Initializaing and checking that the version is supported
        """
        _assert_version(version)
        self.version = version
        self.meta = MODELS[version](**data)

    @classmethod
    def from_file(cls, file_name: str, version: str) -> "MAML":
        """
        Creates a new maml object from file.
        """
        _assert_version(version)
        data = read_maml(file_name)
        return cls(data, version)

    def to_dict(self, include_none: bool = True) -> dict:
        """
        Returns dictionary represenation of the model base class.
        """
        raw = self.meta.model_dump(mode="json")
        if not include_none:
            return _remove_nones(raw)
        return raw

    def to_file(self, file_name: str) -> dict:
        """
        Creates a dictionary representation of the base model.
        """
        SafeDumper.add_representer(
            type(None),
            lambda dumper, _: dumper.represent_scalar("tag:yaml.org,2002:null", ""),
        )
        root, ext = os.path.splitext(file_name)
        if ext != ".maml":
            raise ValueError(f"Extension '{ext}' is not a valid maml extension.")
        with open(f"{root}.maml", "w", encoding="utf8") as file:
            yaml.safe_dump(
                self.to_dict(), file, sort_keys=False, default_flow_style=False
            )


class MAMLBuilder:
    """
    Builder pattern for constructing the MAML format based on whatever version is decided.
    """

    def __init__(self, version: str):
        """
        Initializing and checking version is valid.
        """
        _assert_version(version)
        self.version: str = version
        self._model_cls = MODELS[version]
        self._data: dict = {}

    def set(self, field, value):
        """
        For setting scalar values
        """
        self._data[field] = value
        return self

    def add(self, field, value):
        """
        For adding vector values
        """
        self._data.setdefault(field, []).append(value)
        return self

    def build(self):
        """
        Attempts to build the class for the current version
        """
        return MAML(self._data, self.version)



"""
Helper module to parse and check valid maml data structures.
"""

import datetime

from pydantic_core import ValidationError

from .model_v1p0 import V1P0
from .model_v1p1 import V1P1
from .read import read_maml


MODELS = {
    "v1.0": V1P0,
    "v1.1": V1P1,
}


def today() -> str:
    """
    Returns todays date in the correct iso format.
    """
    return datetime.date.isoformat(datetime.date.today())


def is_iso8601(date: str) -> bool:
    """
    Validates that the given date is in the ISO 8601 format (https://en.wikipedia.org/wiki/ISO_8601)
    """
    try:
        datetime.datetime.fromisoformat(date)
        return True
    except ValueError:
        return False


def _assert_version(version: str) -> None:
    """
    Determines if the version is supported and crashes if it isn't.
    """
    if version not in MODELS:
        raise ValueError(
            f"{version} is not a valid version. Supported MAML versions: {list(MODELS.keys())}"
        )


def valid_for(file_name: str) -> list[str]:
    """
    Reads in a file and determines if it is valid for versions of maml.
    """
    dictionary = read_maml(file_name)
    valid = []
    for version, model in MODELS.items():
        try:
            model(**dictionary)
            valid.append(version)
        except ValidationError:
            pass
    if not valid:
        return ["Not valid for any version of MAML"]
    return valid
